/*
	For best results in visual layout while viewing this file, set
	tab stops to every 4 columns.
*/

/*
	mail.c - Mailer User-Agent (UA)

	version	1.0	Stuart Lynne
	version 1.5	Samuel Lam <skl@van-bc.UUCP>	August/87
*/

#include <stdio.h>

#include "lib.h"
#include "mlib.h"
#include "host.h"

#include "getopt.h"

extern void lmail();

#define LSIZE	256

char mfilename[132], thefilename[50], tmailbox[64], tmailbag[64];
FILE *fmailbox, *rmailbox, *fmailbag;

static int PrintOnly = FALSE;

static int letternum;	/* letters in the mailbox */
static int modified, PageCount;

#define MAXLETTERS	300

static struct ldesc {
	int deleted;	/* status of this message	*/
	long adr;		/* address of From line		*/
	long date;		/* address of Date: line	*/
	long subject;	/* address of Subject: line	*/
	long from;		/* address of From: line	*/
	long lines;		/* number of lines			*/
} letters[MAXLETTERS];

void Cleanup(), Interactive_Mail(), PrintSubject(), UpdateMailbox();
void CopyMsg(), Pager(), PageReset(), Collect_Mail();

int PageLine();
char *NextToken();

static char	*Help_Message = "\
d\tdelete current message\n\
dq\tdelete current message and then quit\n\
f\tmove forward one message\n\
f user\tforward current message to user\n\
g num\tmake message num current message\n\
h\tprint all subjects\n\
H num\tprint subject of current message\n\
m user\tmail a message to user\n\
n\tmove forward one message (no delete)\n\
p\tprint current message\n\
pp\tprint current message (use built-in pager)\n\
q\tquit with mailbox updated\n\
s[file]\tsave current message to file(s) (default =mbox)\n\
u\tundelete current message\n\
w[file]\tsave current message to file(s) without header (default =mbox)\n\
x\texit with mailbox restored\n\
y[file]\tsave current message to file(s) (default =mbox)\n\
-\tmove backward one message (no delete)\n\
+\tmove forward one message (no delete)\n\
^\tmove backward one message (no delete)\n\
<num>\tmake message <num> the current message";


int mailmain(argc, argv)
int argc;
char *argv[];
{
	debuglevel = 0;
	logfile = stdout;
	logecho = FALSE;

	mkfilename(tmailbox, tempdir, "mailbox");
	mkfilename(tmailbag, tempdir, "mailbag");

	unlink(tmailbox);

	if (argc == 1 || (argv[1][0] == '-') && (argv[1][1] != 's'))
		Interactive_Mail(argc, argv);
	else
		Collect_Mail(argc, argv);

	Cleanup();

	return 0;

} /*mailmain*/


void Cleanup()
{

	if (fmailbag != nil(FILE))
		fclose(fmailbag);
	unlink(tmailbox);
	unlink(tmailbag);

} /*Cleanup*/


/*
	Send_Mail

	Send text in a mailbag file to address(es) specified by line.
*/

void Send_Mail(mailbag, argc, argv)
char *mailbag;
int argc;
char *argv[];
{

	/* append user's signature file, if there is one */
	if (E_signature != nil(char)) {
		char sigfile[100], buf[BUFSIZ];
		FILE *sigfp;
		mkfilename(sigfile, homedir, E_signature);
		printmsg(4, "Send_Mail: signature file %s", sigfile);
		if ((sigfp = FOPEN(sigfile, "r", TEXT)) != nil(FILE)) {
			FILE *mailbag_fp;
			mailbag_fp = FOPEN(mailbag, "a", TEXT);
			fputs("-- \n", mailbag_fp);
			while (fgets(buf, BUFSIZ, sigfp) != nil(char))
				fputs(buf, mailbag_fp);
			fclose(sigfp);
			fclose(mailbag_fp);
		} else
			printmsg(0, "signature file \"%s\" doesn't exist!", sigfile);
	}

	/* use mailbag as stdin for mail delivery */
	FILEMODE(TEXT);
	if (freopen(mailbag, "r", stdin) != nil(FILE))
		lmail(argc, argv);

	FILEMODE(TEXT);
	freopen(CONSOLE, "r", stdin);

	/* get rid of the mailbag */
	unlink(mailbag);

} /*Send_Mail*/


void Interactive_Mail(argc, argv)
int argc;
char *argv[];
{
	char line[LSIZE], resp[LSIZE];
	int error, i, inheader, option;
	struct ldesc *letter;

	/* get mailbox file name */
	mkfilename(mfilename, maildir, mailbox);

	/* parse arguments */
	while ((option = getopt(argc, argv, "f:px:")) != EOF)
		switch (option) {
		case 'f':
			if (*optarg == '=')		/* relative to home directory? */
				mkfilename(mfilename, homedir, optarg + 1);
			else
				strcpy(mfilename, optarg);
			break;
		case 'p':
			PrintOnly = TRUE;
			break;
		case 'x':
			debuglevel = atoi(optarg);
			break;
		case '?':
			puts("\nUsage:\tmail [-s subject] recipient ...");
			puts("\tmail [-f mailbox] [-p] [-b]");
			return;
		}

	argv = &argv[optind];

	/* open real mailbox file */
	if ((rmailbox = FOPEN(mfilename, "r", TEXT)) == nil(FILE)) {
		printf("No mail in %s\n", mfilename);
		return;
	}

	/* open up tmp mailbox file */
	if ((fmailbox = FOPEN(tmailbox, "w", BINARY)) == nil(FILE)) {
		printf("cannot open %s for writing\n", tmailbox);
		return;
	}

	/* copy real mailbox file to tmp mailbox file */
	letternum = 0;
	inheader = FALSE;
	while (fgets(line, LSIZE, rmailbox) != nil(char)) {

		if (equaln(line, "From ", 5)) {
			letter = &letters[letternum++];
			letter->from = letter->subject = letter->date = -1L;
			letter->adr = ftell(fmailbox);
			letter->deleted = FALSE;
			letter->lines = 0L;
			inheader = TRUE;
		} else
			if (inheader)
				if (*line == '\n')
					inheader = FALSE;
				else if (equaln(line, "Date:", 5))
					letter->date = ftell(fmailbox);
				else if (equaln(line, "From:", 5))
					letter->from = ftell(fmailbox);
				else if (equaln(line, "Subject:", 8))
					letter->subject = ftell(fmailbox);
				else ;
			else
				letter->lines++;

		fputs(line, fmailbox);

	}

	letters[letternum].adr = ftell(fmailbox);

	/* close mailbox files */
	fclose(rmailbox);
	fclose(fmailbox);

	fmailbox = FOPEN(tmailbox, "r", BINARY);

	modified = FALSE;

	if (PrintOnly) {
		int j = 0;
		while (j < letternum)
			Pager(j++, TRUE);
		return;
	}

	PrintSubject(-1);	/* print all subjects */

	for (i = 0; ; ) {
		char *cp, *p;

		if (letternum == 0)
			i = -1;
		else
			if (i < 0)
				i = 0;
		if (i >= letternum)
			i = letternum - 1;

		printf("%d", i + 1);

		if (!Console_fgets(resp, LSIZE, " ? "))
			break;

		if (*(cp = resp + strlen(resp) - 1) == '\n')
			*cp = '\0';	/* zap any trailing \n */

		switch (*resp) {

		default:	/* invalid commands */
			puts("Usage:");

		case '?':	/* help */
			puts(Help_Message);
			break;

		case '+':	/* move forward one message */
		case 'n':
			i++;
			break;

		case 'g':	/* goto a message */
			i = atoi(resp + 2) - 1;
			break;

		case '0':	/* goto a message */
		case '1':
		case '2':
		case '3':
		case '4':
		case '5':
		case '6':
		case '7':
		case '8':
		case '9':
			i = atoi(resp) - 1;
			break;

		case 'x':	/* exit without updating mailbox */
			modified = FALSE;
			/* fall through */

		case 'q':	/* exit and update mailbox */
			goto donep;

		case '\0':	/* null command */
			break;

		case 'p':	/* print current message */
			if (letternum == 0) {
				puts("mailbox empty");
				continue;
			}
			Pager(i, resp[1] == 'p');
			break;

		case '^':	/* move backward one message */
		case '-':
			i--;
			break;

		case 's':	/* save or write current message to file */
		case 'y':
		case 'w':

			if (resp[1] != '\0' && resp[1] != ' ') {
				puts("Illegal file name");
				continue;
			}
			if (letternum == 0) {
				puts("Mailbox empty");
				continue;
			}

			/* default the file name */
			if (resp[1] == '\0')
				strcat(resp, " =mbox");

			error = FALSE;
			for (p = resp + 1; (p = NextToken(&cp, p)) != nil(char); ) {
				if (*cp == '=')		/* relative to home directory? */
					mkfilename(thefilename, homedir, ++cp);
				else
					strcpy(thefilename, cp);
				printf("%s to %s\n",
					(resp[0] == 's') ? "Saving" : "Writing", thefilename);
				if ((fmailbag = FOPEN(thefilename, "a", TEXT)) == nil(FILE)) {
					printf("cannot append to %s\n", thefilename);
					error = TRUE;
					continue;
				}
				CopyMsg(i, fmailbag, (resp[0] != 'w'));
				fclose(fmailbag);
			}

			if (!error) {
				letters[i].deleted = TRUE;
				modified = TRUE;
				i++;
			}

			break;

		case 'm':	/* compose and send a message */

			if (resp[1] != ' ') {
				puts("invalid mail command");
				continue;
			}

			/* collect the message and send it out */
			{
				char *Largv[50];
				int	Largc;
				Largv[0] = "lmail";
				Largc = getargs(resp + 2, Largv + 1);
				Collect_Mail(Largc + 1, Largv);
				/* can't tell if mailing had worked (yet) */
			}

			break;

		case 'f':	/* forward message *or* move forward */

			if (resp[1] == '\0') {
				i++;	/* move forward one message */
				continue;
			}

			if (resp[1] != ' ') {
				puts("Invalid forward command.");
				continue;
			}
			if (letternum == 0) {
				puts("Mailbox empty.");
				continue;
			}

			/* copy current message to a temporary file */
			fmailbag = FOPEN(tmailbag, "w", TEXT);
			CopyMsg(i, fmailbag, TRUE);
			fclose(fmailbag);

			/* mail the content of the temporary file */
			{
				char	*Largv[50];
				int	Largc;
				Largv[0] = "lmail";
				Largc = getargs(resp + 2, Largv + 1);
				Send_Mail(tmailbag, Largc + 1, Largv);
			}

			/* no way to tell if forwarding is successful yet */
			error = TRUE;
			if (!error) {
				letters[i].deleted = TRUE;
				modified = TRUE;
			}
			i++;

			break;

		case 'd':	/* delete current message */
			letters[i].deleted = TRUE;
			modified = TRUE;
			i++;
			if (resp[1] == 'q')	/* the 'dq' command? */
				goto donep;
			break;

		case 'h':	/* print all messages' subjects */
			PrintSubject(-1);
			break;

		case 'H':	/* print current message's subject */
			PrintSubject(i);
			break;

		case 'u':	/* undelete current message */
			letters[i].deleted = FALSE;
			break;

		}

	}
donep:

	if (modified)
		UpdateMailbox();

} /*Interactive_Mail*/


int RetrieveLine(adr, line)
long adr;
char *line;
{
	char buf[132];
	char *cp;

	*line = '\0';
	if ((adr == -1L) || (fseek(fmailbox, adr, SEEK_SET) != 0) ||
	   (fgets(buf, sizeof(buf), fmailbox) == nil(char)))
		return FALSE;

	strcpy(line, buf);

	if ((cp = strchr(line, '\n')) != nil(char))
		*cp = '\0';

	return TRUE;

} /*RetrieveLine*/


#if FALSE

/* The follwing procedure (ReturnAddress) does not work yet! */

void ReturnAddress(line)
char *line;
{
	char *i, *k;
	(void)RetrieveLine(lp->from, line);
	if (
		(i = strchr(line, '<')) != null(char) &&
		(k = strchr(line, '>')) != null(char) &&
		(k > i)
		)
	{
		k -= i;
		strncpy(from, line[i], k);
		from[k] = '\0';
	}
	else {
	}
} /*ReturnAddress*/

#endif	/*FALSE*/


void PrintSubject(msgnum)
int msgnum;
{
	struct ldesc *ld;
	char from[132], subject[132], date[132], line[132];
	int k, mink, maxk;

	PageReset();

	if (msgnum == -1) {		/* print all of them? */
		mink = 0;
		maxk = letternum - 1;
	} else
		mink = maxk = msgnum;

	for (k = mink ; k <= maxk ; k++) {

		ld = &letters[k];

		strcpy(from, "unknown");	/* default to "unknown" */
		if (RetrieveLine(ld->from, line)) {
			register char *left, *right, *start, *dest;

			/* Assume From: mailbox@site.domain */

			/* get first usable stuff after From: */
			start = strchr(line, ':') + 1;

			/* do we have From: mailbox@site.domain (Proper Name) */
			if (
				((left = strchr(line, '(')) != nil(char)) &&
				((right = strchr(line, ')')) != nil(char))
				)
				start = left + 1;
			else
				/* or we may have it in the form
				   From: Proper Name <mailbox@site.domain> */
				right = strchr(line, '<');

			if (*right == '<' || *right == ')')
				right--;

			/* if we didn't get either of the last two, then
			   just grab the line */
			if (right < start || right == nil(char)) {
				char *p, *p1, *p2;
				/* point to last character */
				right = line + strlen(line) - 1;
				/* try optimize a!b!...!y!z to just y!z */
				p1 = p2 = nil(char);
				p = start;
				while ((p = strchr(p, '!')) != nil(char)) {
					p2 = p1;
					p1 = p++;
				}
				if ((p1 != nil(char)) && (p2 != nil(char)))
					start = p2 + 1;
			}

			/* trim leading blanks */
			while (isspace(*start))
				start++;

			/* trim trailing blanks */
			while (isspace(*right))
				right--;

			/* finally copy the address over */
			dest = from;
			while (start <= right)
				*dest++ = *start++;
			*dest = '\0';
		}

		/* Date: Wed May 13 23:59:53 1987 */
		*date = '\0';	/* default date to null */
		if (RetrieveLine(ld->date, date)) {
			sscanf(date, "%*s %*s %s %s", line, subject);
			sprintf(date, "%s %s", line, subject);
		}

		strcpy(subject, "--- no subject ---");
		if (RetrieveLine(ld->subject, line)) {
			register char	*sp;
			sp = line;
			while (!isspace(*sp))
				sp++;
			while (isspace(*sp))
				sp++;
			strcpy(subject, sp);
		}

		/* make sure the fields aren't too long */
		from[25] = '\0';
		date[6] = '\0';
		subject[30] = '\0';

		sprintf(line, "%3d%c %6s  %-25s  %-30s  (%5ld)\n", k + 1,
			(ld->deleted ? '*' : ' '), date, from, subject,
			ld->lines);

		if (PageLine(line))
			break;

	}

} /*PrintSubject*/


void UpdateMailbox()
{
	register int i;
	int changed = FALSE;

	for (i = 0; i < letternum; i++)
		if (letters[i].deleted) {
			changed = TRUE;
			break;
		}
	if (!changed)
		return;

	if ((fmailbag = FOPEN(mfilename, "w", TEXT)) == nil(FILE)) {
		printf("UpdateMailbox: can't rewrite %s.\n", mfilename);
		Cleanup();
	}
	for (i = 0; i < letternum; i++)
		if (!(letters[i].deleted)) {
			CopyMsg(i, fmailbag, TRUE);
		}
	fclose(fmailbag);

} /*UpdateMailbox*/


/*
   CopyMsg - Copy a message

   If noheaders is TRUE, don't copy the header lines over.
*/

void CopyMsg(msgnum, f, headers)
int msgnum;
FILE *f;
int headers;			/* include headers? */
{
	long nextloc;
	char buf[BUFSIZ];

	fseek(fmailbox, letters[msgnum].adr, SEEK_SET);
	nextloc = letters[msgnum + 1].adr;

	while (ftell(fmailbox) < nextloc &&
		fgets(buf, BUFSIZ, fmailbox) != nil(char)) {

		/* write line */
		if (headers)
			fputs(buf, f);

		/* reset header */
		if (!headers && equal(buf, "\n"))
			headers = TRUE;

	} /*while*/

} /*CopyMsg*/


/*
	Pager - page through a message

	There are hooks here to let the user use his/her own pager,
	like LIST, MORE, or LESS.  We just write the message out to
	a temporary file and invoke the appropriate external program
	to do the browsing.
*/

void Pager(msgnum, builtin)
int msgnum;
int builtin;
{
	long nextloc;
	char buf[BUFSIZ];

	if (msgnum == -1)
		return;

	if (!builtin && (*E_pager != '\0')) {

		if ((fmailbag = FOPEN(tmailbag, "w", TEXT)) == nil(FILE)) {
			puts("cannot open browse file");
			return;
		}
		CopyMsg(msgnum, fmailbag, TRUE);
		fclose(fmailbag);

		L_invoke_pager(E_pager, tmailbag);

	} else {

		fseek(fmailbox, letters[msgnum].adr, SEEK_SET);
		nextloc = letters[msgnum + 1].adr;

		PageReset();
		while (ftell(fmailbox) < nextloc &&
			fgets(buf, BUFSIZ, fmailbox) != nil(char))
			if (PageLine(buf))
				break;
		PageLine("\n");

	}

} /*Pager*/


void PageReset()
{

	PageCount = 0;

} /*PageReset*/


int	PageLine(line)
char *line;
{

	fputs(line, stdout);

	if (++PageCount > 22) {
		PageCount = 0;
		fputs("More?", stdout);
		switch (tolower(Get_One())) {
		case 'q':
		case '\003':
		case 'x':
			puts("\r     \nAborted.\n");
			return TRUE;
		default:
			printf("\r     \r");
		}
	}

	return FALSE;

} /*PageLine*/


char *NextToken(token, start)
char **token, *start;
{
	register char *p = start;

	*token = nil(char);

	/* skip leading blanks and tabs */
	while (*p == ' ' || *p == '\t')
		p++;

	if (*p == '\0')
		return nil(char);

	*token = p;

	/* find end of token */
	while (*p != ' ' && *p != '\t' && *p != '\0')
		p++;

	if (*p != '\0')
		*p++ = '\0';

	return p;

} /*NextToken*/


void Collect_Mail(argc, argv)
int argc;
char *argv[];
{
	int Interactive;
	char buf[BUFSIZ];

	Interactive = Is_Console(stdin);

	if (Interactive)
		puts("\nEnter message.  (End input with end-of-file.)");

	/* copy a message from stdin to temporary file */
	fmailbag = FOPEN(tmailbag, "w", TEXT);
	for ( ; ; ) {
		if (Interactive)
			if (!Console_fgets(buf, BUFSIZ, "? "))
				break;
			else ;
		else
			if (fgets(buf, BUFSIZ, stdin) == nil(char))
				break;
		fputs(buf, fmailbag);
		if (buf[strlen(buf)-1] != '\n')
			fputc('\n', fmailbag);
	}
	fclose(fmailbag);

	if (Interactive) {
		int done;
		do {
			fputs("\nSend, Edit, Abort? ", stdout);
			switch (tolower(Get_One())) {
			case 's':
				puts("Send");
				Send_Mail(tmailbag, argc, argv);
				done = TRUE;
				break;
			case 'e':
				puts("Edit");
				Invoke_Editor(E_editor, tmailbag);
				done = FALSE;
				break;
			case 'a':
				fputs("Abort\nAre you sure? ", stdout);
				switch (tolower(Get_One())) {
				case 'y':
					puts("Yes");
					done = TRUE;
					break;
				default:
					puts("No");
					done = FALSE;
				} /*switch*/
				break;
			default:
				puts("Enter S, E or A.");
				done = FALSE;
			} /*switch*/
		} while (!done);
	} else
		Send_Mail(tmailbag, argc, argv);

} /*Collect_Mail*/